@@ -1,4 +1,5 @@
 from django.db import models
+from django.db.models.query import EmptyQuerySet
 
 from neo4django.decorators import not_implemented
 from query import NodeQuerySet
@@ -7,6 +8,7 @@ class NodeModelManager(models.Manager):
     def __init__(self):
         self._using = None
         self.model = None
+        self._inherited = False
 
     @not_implemented
     def _insert(self, values, **kwargs):
@@ -16,8 +18,11 @@ class NodeModelManager(models.Manager):
     def _update(self, values, **kwargs):
         pass   
 
-    @not_implemented
     def get_empty_query_set(self):
+        return EmptyQuerySet()
+
+    @not_implemented
+    def exclude(self, *args, **kwargs):
         pass
 
     def get_query_set(self):
@@ -36,4 +41,6 @@ class NodeModelManager(models.Manager):
         return self.get_query_set().create(**kwargs)
     
     def filter(self, *args, **kwargs):
+        if args:
+            raise NotImplementedError('The Q operator is not currently supported')
         return self.get_query_set().filter(**kwargs)
