@@ -18,7 +18,9 @@ from .. import connections
 from neo4django.validators import validate_array, validate_str_array,\
         validate_int_array, ElementValidator
 from neo4django.utils import AttrRouter, write_through
-from neo4django.constants import AUTO_PROP_INDEX_VALUE
+from neo4django.constants import ERROR_ATTR
+
+from dateutil.tz import tzutc, tzoffset
 
 MIN_INT = -9223372036854775808
 MAX_INT = 9223372036854775807
@@ -41,7 +43,7 @@ class Property(object):
                  has_own_index=False, unique=False, name=None, editable=True,
                  null=True, blank=True, validators=[], choices=None,
                  error_messages=None, required=False, serialize=True,
-                 auto=False, metadata={}, auto_default=NOT_PROVIDED, 
+                 auto=False, metadata={}, auto_default=NOT_PROVIDED,
                  default=NOT_PROVIDED, **kwargs):
         if unique and not indexed:
             raise ValueError('A unique property must be indexed.')
@@ -195,9 +197,9 @@ class Property(object):
         and validate are propagated. The correct value is returned if no error is
         raised.
         """
-        value = self.to_python(value)
         self.validate(value, model_instance)
         self.run_validators(value)
+        value = self.to_python(value)
         return value
 
     def pre_save(self, model_instance, add, attname):
@@ -221,6 +223,7 @@ class BoundProperty(AttrRouter):
                      'unique',
                      'to_neo',
                      'to_neo_index',
+                     'to_neo_index_gremlin',
                      'member_to_neo_index',
                      'from_python',
                      'from_neo',
@@ -258,15 +261,9 @@ class BoundProperty(AttrRouter):
         return cmp(self.creation_counter, other.creation_counter)
 
     @staticmethod
-    def __values_of(instance, create=True):
+    def _values_of(instance, create=True):
         try:
             values = instance._prop_values
-            #for key, prop in BoundProperty._all_properties_for(instance).items(): #XXX: Might be a faster/more elegant way
-            #    if prop.attname not in values:
-            #        if getattr(prop._property, 'auto_now', False):
-            #            values[prop.__attname] = datetime.datetime.now() #XXX:Setting to None here sets the node's datetime to None
-            #            if type(prop._property) == DateProperty: #XXX:Kinda gross way to handle it :\
-            #                values[prop.__attname] = datetime.datetime.date(values[prop.__attname])
         except:
             values = {}
             if create:
@@ -301,76 +298,91 @@ class BoundProperty(AttrRouter):
         else:
             return self.__class.index(using)
 
-    def _save_(instance, node, node_is_new): #TODO this entire method could be transactional
-        values = BoundProperty.__values_of(instance)
+    def index_name(self, using):
+        if not (self.indexed or self.auto):
+            raise TypeError("'%s' is not indexed" % (self.__propname,))
+        else:
+            return self.__class.index_name(using)
+
+    #update the state of the model instance based on a rest client element property dictionary
+    def _update_values_from_dict(instance, new_val_dict, clear=False):
+        values = BoundProperty._values_of(instance)
         properties = BoundProperty._all_properties_for(instance)
+
+        values.clear()
+        
+        for k, v in new_val_dict.items():
+            prop = properties.get(k, None)
+            if prop:
+                #XXX duplicates __get_value()...
+                values[k] = prop.from_neo(v)
+        #XXX this relies on neo4jrestclient private implementation details
+        if clear:
+            instance.node._dic['data'].clear()
+        instance.node._dic['data'].update(new_val_dict)
+
+    NodeModel._update_values_from_dict = staticmethod(_update_values_from_dict) #TODO this needs to be revised
+    del _update_values_from_dict
+
+    def _save_(instance, node, node_is_new):
+        values = BoundProperty._values_of(instance)
+        properties = BoundProperty._all_properties_for(instance)
+
+        gremlin_props = {}
         for key, prop in properties.items():
-            index = None
+            prop_class = prop.__class
+            prop_dict = gremlin_props[key] = {}
             if prop.auto and values.get(key, None) is None:
-                type_node = prop.target._type_node(instance.using)
-                
-                last_auto_attname = '%s.%s' % (prop.attname, AUTO_PROP_INDEX_VALUE)
-
-                script = prop.next_value_gremlin
-                script += \
-                """
-                typeNode = g.v(typeNodeID)
-                rawTypeNode = typeNode.getRawVertex()
-
-                lockManager.getWriteLock(rawTypeNode)
-
-                if (lastAutoProp in typeNode.map) {
-                    value = typeNode[lastAutoProp]
-                    value = nextValue(value)
-                }
-                else {
-                    value = defaultAutoVal
-                }
-                typeNode[lastAutoProp] = value
-
-                lockManager.releaseWriteLock(rawTypeNode, null)
-                
-                results = value
-                """
-                conn = connections[instance.using]
-                value = conn.gremlin_tx_deadlock_proof(script, 0,
-                      defaultAutoVal=prop.auto_default, lastAutoProp=last_auto_attname,
-                      typeNodeID=type_node.id)
-                #if the code is interrupted between here and setting
-                #the value, there might be a problem... worst case, lost
-                #id space?
-
-                values[key] = value
+                prop_dict['auto_increment'] = True
+                prop_dict['increment_func'] = prop.next_value_gremlin
+                prop_dict['auto_default'] = prop.auto_default
+                prop_dict['auto_abstract'] = prop_class._meta.abstract
+                prop_dict['auto_app_label'] = prop_class._meta.app_label
+                prop_dict['auto_model'] = prop_class.__name__
+            if prop.indexed:
+                prop_dict['index_name'] = prop.index_name(instance.using)
+                if hasattr(prop, 'to_neo_index_gremlin'):
+                    prop_dict['to_index_func'] = prop.to_neo_index_gremlin
             if key in values:
                 value = values[key]
+                prop.clean(value, instance)
                 value = prop.pre_save(node, node_is_new, prop.name) or value
+                if not value in validators.EMPTY_VALUES or getattr(prop._property, "use_string", False):
+                    #should already have errored if self.null==False
+                    value = prop.to_neo(value)
                 values[key] = value
-                old, value = prop.__set_value(instance, value)
+                prop_dict['value'] = value
                 if prop.indexed:
-                    index = index if index else prop.index(using=instance.using)
-                    if prop.unique:#TODO empty values? in validators.empty? # and value is not None:
-                        try:
-                            old_node = index[prop.attname][value]
-                        except NotFoundError, e:
-                            old_node = None
-                        if old_node and old_node != node:
-                            raise ValueError(
-                                "Duplicate index entries for <%s>.%s" %
-                                (instance.__class__.__name__,
-                                    prop.name))
-                    if old is not None:
-                        index.delete(prop.attname, None, node)
+                    indexed_values = prop_dict['values_to_index'] = []
+                    prop_dict['unique'] = bool(prop.unique)
                     if value is not None:
-                        index.add(prop.attname, prop.to_neo_index(value), node)
+                        indexed_values.append(prop.to_neo_index(value))
                         if prop.indexed_by_member:
                             for m in value:
-                                index.add(prop.attname, prop.member_to_neo_index(m), node)
+                                indexed_values.append(prop.member_to_neo_index(m))
+        script = '''
+        node=g.v(nodeId);
+        results = Neo4Django.updateNodeProperties(node, propMap);
+        '''
+        conn = connections[instance.using]
+        script_rv= conn.gremlin_tx(script, nodeId=instance.id,
+                propMap=gremlin_props, raw=True)
+
+        if isinstance(script_rv, dict) and ERROR_ATTR in script_rv and 'property' in script_rv:
+                raise ValueError( "Duplicate index entries for <%s>.%s" % 
+                                 (instance.__class__.__name__, script_rv['property']))
+        elif isinstance(script_rv, dict) and 'data' in script_rv:
+            #returned a node (TODO #128 error passing generalization)
+            NodeModel._update_values_from_dict(instance, script_rv['data'], clear=True)
+        else:
+            raise ValueError('Unexpected response from server: %s' % str(script_rv))
+        
     NodeModel._save_properties = staticmethod(_save_) #TODO this needs to be revised. I hope there's a better way.
     del _save_
 
     def __get__(self, instance, cls=None):
         if instance is None: return self
-        values = self.__values_of(instance, create=False)
+        values = self._values_of(instance, create=False)
         if self.__propname in values:
             return values[self.__propname]
         else:
@@ -380,7 +392,7 @@ class BoundProperty(AttrRouter):
         if write_through(instance):
             self.___set_value(instance, value)
         else:
-            values = self.__values_of(instance)
+            values = self._values_of(instance)
             values[self.__propname] = value
 
     @transactional
@@ -391,7 +403,7 @@ class BoundProperty(AttrRouter):
             pass
         else:
             try:
-                values = BoundProperty.__values_of(instance)
+                values = BoundProperty._values_of(instance)
                 values[self.__propname] = val = self._property.from_neo(underlying[self.__propname])
                 return val
             except: # no value set on node
@@ -489,6 +501,15 @@ class IntegerProperty(Property):
             raise ValueError('Values should be between {0} and {1}.'.format(MIN_INT, MAX_INT))
         return ('-' if value < 0 else '0') + s.zfill(19)
 
+    @property
+    def to_neo_index_gremlin(self):
+        """
+        Return a Gremlin/Groovy closure literal that can compute 
+        to_neo_index(value) server-side. The closure should take a single value
+        as an argument (that value actually set on the node).
+        """
+        return """{ i -> (i < 0?'-':'0') + String.format('%019d',i)} """
+
 class AutoProperty(IntegerProperty):
     def __init__(self, *args, **kwargs):
         kwargs['auto'] = True
@@ -504,15 +525,11 @@ class AutoProperty(IntegerProperty):
     @property
     def next_value_gremlin(self):
         """
-        Return a Gremlin/Groovy closure that can compute next_value()
-        server-side. The function should be named 'nextValue', and take a
-        single value as an argument to increment.
+        Return a Gremlin/Groovy closure literal that can compute next_value()
+        server-side. The closure take a single value as an argument to
+        increment.
         """
-        script = \
-        """
-        def nextValue = { i -> i + 1}
-        """
-        return script
+        return """{ i -> i + 1}"""
 
 class DateProperty(Property):
     __format = '%Y-%m-%d'
@@ -597,7 +614,7 @@ class DateProperty(Property):
             return super(DateProperty, self).pre_save(model_instance, add, attname)
 
 class DateTimeProperty(DateProperty):
-    __format = '%Y-%m-%d-%H:%M:%S.%f'
+    __format = '%Y-%m-%d %H:%M:%S.%f'
 
     default_error_messages = {
         'invalid': _(u'Enter a valid date/time in YYYY-MM-DD HH:MM[:ss[.uuuuuu]] format.'),
@@ -607,8 +624,10 @@ class DateTimeProperty(DateProperty):
     MIN=datetime.datetime.min
 
     @classmethod
-    def _format_datetime(cls, value):
-        time_string = cls._format_date(value, cls.__format)
+    def _format_datetime(cls, value, format_string=None):
+        if not format_string:
+            format_string = cls.__format
+        time_string = cls._format_date(value, format_string)
         return time_string.replace('%H', str(value.hour).zfill(2))\
                           .replace('%M', str(value.minute).zfill(2))\
                           .replace('%S', str(value.second).zfill(2))\
@@ -657,6 +676,9 @@ class DateTimeProperty(DateProperty):
 
         return self._format_datetime(result)
 
+    def to_neo_index(self, value):
+        return self.to_neo(value).replace(' ','-')
+
     def pre_save(self, model_instance, add, attname):
         if self.auto_now or (self.auto_now_add and add):
             value = datetime.datetime.now()
@@ -665,6 +687,108 @@ class DateTimeProperty(DateProperty):
         else:
             return super(DateTimeProperty, self).pre_save(model_instance, add, attname)
 
+class DateTimeTZProperty(DateTimeProperty):
+    '''
+    DateTimeProperty that can store and retrieve timezone-aware datetimes.
+    '''
+    __format = '%Y-%m-%d %H:%M:%S.%f %z'
+
+    @classmethod
+    def _format_offset(cls, offset_timedelta):
+        '''
+        Produce a timezone offset string (+/- HHMM) from a timedelta.
+        '''
+        try:
+            seconds = offset_timedelta.total_seconds()
+        except AttributeError:
+            # total_seconds method is only available from 2.7 up
+            td = offset_timedelta
+            days_to_secs = td.days * 24 * 3600.0
+            secs_to_micro = (td.seconds + days_to_secs) * (10 ** 6)
+            seconds = (td.microseconds + secs_to_micro) / (10 ** 6)
+        mins = seconds / 60
+        hrs = mins / 60
+        mins = mins % 60
+        return '%+03d%02d' % (hrs, mins)
+
+    @classmethod
+    def _parse_tz(cls, tz_str=None):
+        '''
+        Read a timezone string of the form '+0000' and return a timezone
+        object. If not given, just return UTC.
+        '''
+        tz_str = tz_str.strip() if tz_str else ''  # Ensure no whitespace
+        if (not tz_str) or (tz_str == '+0000') or (tz_str == '-0000'):
+            # Shortcut for the common case where it's UTC, or default
+            return tzutc()
+        # Otherwise, pull out the hours and minutes and construct a
+        # tzoffset(), which requires an offset in seconds
+        hrs = int(tz_str[1:3])
+        mins = int(tz_str[3:5])
+        mult = -1 if (tz_str[0] == '-') else 1
+        offset = mult * ((hrs * 3600) + (mins * 60))
+        return tzoffset('LOCAL', offset)
+
+    @classmethod
+    def _format_datetime_with_tz(cls, value):
+        '''
+        Format a datetime (e.g. for storage in Neo4j) with a timezone offset
+        appended as +/- HHMM.
+        '''
+        formatted = cls._format_datetime(value, cls.__format)
+        if value.utcoffset() is not None:
+            offset_string = cls._format_offset(value.utcoffset())
+        else:
+            offset_string = ""
+        return formatted.replace("%z", offset_string).strip()
+
+    @classmethod
+    def __parse_datetime_string_with_tz(cls, value):
+        '''
+        Parse a stringified datetime into a datetime object, first trying to
+        read a timezone (if one is provided in our format). Uses the superclass
+        method to parse the actual string, and adds any timezone information
+        at the end.
+        '''
+        try:
+            # Try converting with timezone offset. Since strptime is decidedly
+            # inconsistent with support for '%z', this must be done manually:
+            # if a '+HHMM' is present, it'll form the last five characters of
+            # the string
+            dt_val, tz_str = value[:-5], value[-5:]
+            dt_val = dt_val.strip()  # ensure no trailing whitespace
+            tz_info = cls._parse_tz(tz_str)
+        except ValueError:
+            tz_info = None
+            dt_val = value
+        # HACK: Relies on CPython 2.x's double-underscore name-mangling!
+        dt = cls._DateTimeProperty__parse_datetime_string(dt_val)
+        return dt.replace(tzinfo=tz_info)
+
+    def from_neo(self, value):
+        if value is None or value == '':
+            return None
+        if isinstance(value, datetime.datetime):
+            return value
+        if isinstance(value, datetime.date):
+            return datetime.datetime(value.year, value.month, value.day)
+
+        return self.__parse_datetime_string_with_tz(value)
+
+    def to_neo(self, value):
+        result = None
+
+        if value is None:
+            return ''
+        if isinstance(value, datetime.datetime):
+            result = value
+        elif isinstance(value, datetime.date):
+            result = datetime.datetime(value.year, value.month, value.day)
+        else:
+            result = self.__parse_datetime_string_with_tz(value)
+
+        return self._format_datetime_with_tz(result)
+
 class ArrayProperty(Property):
     __metaclass__ = ABCMeta
 
@@ -693,14 +817,42 @@ class ArrayProperty(Property):
                 el_val = ElementValidator(vals_or_tuple)
             self.validators.append(el_val)
 
-    def get_default(self):
-        return []
+        #Store array values as a token separated string. For use in the event
+        #the user needs to access the neo4j data multiple ways.
+        #For example using REST interface you cannot store an empty array
+        self.use_string = kwargs.get("use_string", False)
+        self.token = kwargs.get("token", ",")
+        self.escape_token = kwargs.get("escape_token", "+")
+        self.token_regex = re.compile(
+            "(?<!%s)%s" % (re.escape(self.escape_token), self.token))
 
-    #def from_neo(self, value):
-    #    from nose.tools import set_trace; set_trace()
-    #    if value is None:
-    #        return []
-    #    return value
+    def get_default(self):
+        if self.use_string:
+            return ""
+        else:
+            return []
+
+    def from_neo(self, value):
+        if value and not isinstance(value, (tuple, list)) and self.use_string:
+            array_values = self.token_regex.split(value)
+            for i, v in enumerate(array_values):
+                array_values[i] = v.replace(
+                    "%s%s" % (self.escape_token, self.token), self.token)
+            return tuple(array_values)
+        if not value:
+            return tuple([])
+        else:
+            return tuple(value)
+
+    def to_neo(self, value):
+        if self.use_string:
+            escaped_values = []
+            for v in value:
+                escaped_values.append(
+                    str(v).replace(self.token, "%s%s" % (self.escape_token,
+                                                         self.token)))
+            return self.token.join(escaped_values)
+        return value
 
 class StringArrayProperty(ArrayProperty):
     default_validators = [validate_str_array]
