@@ -1,13 +1,15 @@
 import neo4jrestclient.client as neo4j
 import neo4jrestclient.constants as neo_constants
 
-from neo4django.db import DEFAULT_DB_ALIAS, connections
-from neo4django.utils import Enum, uniqify
-from neo4django.constants import ORDER_ATTR
-from neo4django.decorators import transactional, not_supported, alters_data, \
+from .. import DEFAULT_DB_ALIAS, connections
+from ...utils import Enum, uniqify
+from ...constants import ORDER_ATTR
+from ...decorators import transactional, not_supported, alters_data, \
         not_implemented
+from . import script_utils
+from .script_utils import id_from_url
 
-from django.db.models.query import QuerySet
+from django.db.models.query import QuerySet, EmptyQuerySet
 from django.core import exceptions
 from django.db.models.loading import get_model
 
@@ -15,34 +17,40 @@ from lucenequerybuilder import Q
 from jexp import J
 
 from collections import namedtuple
-from operator import and_, itemgetter
+from operator import and_, or_
 import itertools
 import re
 
 #python needs a bijective map... grumble... but a reg enum is fine I guess
 #only including those operators currently being implemented
 OPERATORS = Enum('EXACT', 'LT','LTE','GT','GTE','IN','RANGE','MEMBER','CONTAINS',
-                 'STARTSWITH')
+                 'STARTSWITH', 'MEMBER_IN')
 
-Condition = namedtuple('Condition', ['field','value','operator','negate'])
+ConditionTuple = namedtuple('ConditionTuple', ['field','value','operator','negate'])
+class Condition(ConditionTuple):
+    def __init__(self, *args, **kwargs):
+        if 'value' in kwargs:
+            if isinstance(kwargs['value'], list):
+                kwargs['value'] = tuple(kwargs['value'])
+        else:
+            if len(args) > 1:
+                if isinstance(args[1], list):
+                    args = list(args)
+                    args[1] = tuple(args[1])
+        super(Condition, self).__init__( *args, **kwargs)
 
-QUERY_CHUNK_SIZE = 10
+#TODO move this to settings.py
+QUERY_CHUNK_SIZE = 50
 
 #TODO these should be moved to constants
 TYPE_REL = '<<TYPE>>'
 INSTANCE_REL = '<<INSTANCE>>'
 INTERNAL_RELATIONSHIPS = (TYPE_REL, INSTANCE_REL)
 
-##################
-# UTIL FUNCTIONS #
-##################
 
-def id_from_url(url):
-    from urlparse import urlsplit
-    from posixpath import dirname, basename
-    path = urlsplit(url).path
-    b = basename(path)
-    return int(b if b else dirname(path))
+#TODO move to a util module
+def not_none(it):
+    return itertools.ifilter(None, it)
 
 ######################################
 # IN-PYTHON QUERY CONDITION CHECKING #
@@ -57,8 +65,8 @@ def matches_condition(node, condition):
     #if this is an id field, the value should be the id
     if getattr(field, 'id', None):
         att = node.id
-    elif node.get(field.attname, None):
-        att = node[field.attname]
+    elif node.properties.get(field.attname, None) is not None:
+        att = node.properties[field.attname]
     else:
         att = None
 
@@ -70,6 +78,7 @@ def matches_condition(node, condition):
              (op is OPERATORS.GT and att > val) or \
              (op is OPERATORS.GTE and att >= val) or \
              (op is OPERATORS.IN and att in val) or \
+             (op is OPERATORS.MEMBER_IN and any(a in val for a in att)) or \
              (op is OPERATORS.CONTAINS and val in att) or \
              (op is OPERATORS.STARTSWITH and att.startswith(val))
     if neg:
@@ -99,6 +108,10 @@ def q_from_condition(condition):
         q = Q(attname, '*%s*' % escape_wilds(condition.value), wildcard=True)
     elif condition.operator is OPERATORS.MEMBER:
         q = Q(attname, field.member_to_neo_index(condition.value))
+    elif condition.operator is OPERATORS.IN:
+        q = reduce(or_, (Q(attname, field.to_neo_index(v)) for v in condition.value))
+    elif condition.operator is OPERATORS.MEMBER_IN:
+        q = reduce(or_, (Q(attname, field.member_to_neo_index(v)) for v in condition.value))
     #FIXME OBOE with field.MAX + exrange, not sure it's easy to fix though...
     elif condition.operator in (OPERATORS.GT, OPERATORS.GTE, OPERATORS.LT,
                                 OPERATORS.LTE, OPERATORS.RANGE): 
@@ -142,14 +155,14 @@ def q_from_condition(condition):
         q = ~q
     return q
 
-def filter_expression_from_condition(condition):
+def js_expression_from_condition(condition, js_node):
     field, val, op, neg = condition
     name = field.attname
-    end_node = J('position').endNode()
+
     #XXX assumption that attname is how the prop is stored (will change with
     #issue #30
-    has_prop = end_node.hasProperty(name)
-    get_prop = end_node.getProperty(name)
+    has_prop = js_node.hasProperty(name)
+    get_prop = js_node.getProperty(name)
     if op == OPERATORS.EXACT:
         correct_val = get_prop == val
     elif op == OPERATORS.CONTAINS:
@@ -161,6 +174,11 @@ def filter_expression_from_condition(condition):
             correct_val |= get_prop == v
     elif op == OPERATORS.MEMBER:
         correct_val = get_prop.find(J('function(m){return %s;}' % str(J('m') == val))) == 1
+    elif op == OPERATORS.MEMBER_IN:
+        member_in = J('false')
+        for v in val:
+            member_in |= J('m') == v
+        correct_val = get_prop.find(J('function(m){return %s;}' % str(member_in))) == 1
     elif op == OPERATORS.LT:
         correct_val = get_prop < val
     elif op == OPERATORS.LTE:
@@ -181,6 +199,10 @@ def filter_expression_from_condition(condition):
         filter_exp = ~filter_exp
     return filter_exp
 
+def filter_expression_from_condition(condition):
+    end_node = J('position').endNode()
+    return js_expression_from_condition(condition, end_node)
+    
 def return_filter_from_conditions(conditions):
     exprs = [filter_expression_from_condition(c) for c in conditions]
     #construct an expr to exclude the first node and type nodes. consider
@@ -195,20 +217,6 @@ def return_filter_from_conditions(conditions):
 # QUERY EXECUTION #
 ###################
 
-class LazyNode(neo4j.Node):
-    """
-    A version of the REST client node that doesn't update on init.
-    """
-    def __init__(self, url, dic):
-        self._dont_update = True
-        super(LazyNode, self).__init__(url, create=False)
-        self._dic = dic.copy()
-        self._dont_update = False
-
-    def update(self, *args, **kwargs):
-        if not self._dont_update:
-            super(LazyNode, self).update(*args, **kwargs)
-
 def score_model_rel(field_name, bound_rel):
     """
     Scores a model's bound relationship on how likely it is to be the referrent
@@ -226,22 +234,54 @@ def cypher_rel_str(rel_type, rel_dir):
     out = neo_constants.RELATIONSHIPS_OUT
     return dir_strings[rel_dir==out]%('[:`%s`]' % rel_type)
 
-def cypher_match_from_fields(nodetype, fields):
-    #TODO docstring
-    matches = []
-    for i, f in enumerate(fields):
-        rel_matches = []
+def cypher_from_fields(nodetype, fields):
+    """
+    Generates Cypher MATCH and RETURN expressions from `select_related()` style
+    field strings.
+    """
+    #TODO this function is a great example of why there should be some greater
+    # layer of abstraction between query code and script generation. a first 
+    # step would be to write some CypherPrimitive, CypherList, etc.
+    matches, returns = [], []
+    reqd_fields = (field for i, field in enumerate(fields)
+                   if not any(other_field.startswith(field) and field != other_field
+                              for other_field in fields[i:]))
+
+    for i, field in enumerate(reqd_fields):
+        path_name = 'p%d' % i
+        returns.append(path_name)
+
+        rel_match_components = []
         cur_m = nodetype
-        for step in f.split('__'):
-            candidates_on_models = list(
-                reversed(sorted(s for s in 
-                                ((score_model_rel(step,r),r) for _,r in 
-                                 nodetype._meta._relationships.items()) if s > 0)))
+        for step in field.split('__'):
+            #try to propertly match a model field to the provided field string
+            candidates_on_models = sorted((s for s in ((score_model_rel(step,r),r)
+                for _,r in nodetype._meta._relationships.items()) if s > 0), reverse=True)
+            #TODO provide for the case that there isn't a valid candidate
             choice = candidates_on_models[0]
-            rel_matches.append(cypher_rel_str(choice[1].rel_type, choice[1].direction))
-        matches.append('p%d=(s%d%sr%d)'  %(i, i, '()'.join(rel_matches), i))
-        matches.append('pt%d=(t%d-[:`%s`]->r%d)' % (i, i, INSTANCE_REL, i))
-    return matches
+            rel_match_components.append(
+                cypher_rel_str(choice[1].rel_type, choice[1].direction))
+        
+        node_match_components = [] # Cypher node identifiers
+        type_matches = [] # full Cypher type matching paths for return types
+        for ri in xrange(len(rel_match_components)):
+            return_node_name = '%s_r%d' % (path_name, ri)
+            return_node_type_name = '%s_t' % return_node_name
+
+            returns.extend((return_node_name, '%s.name' % return_node_type_name))
+
+            node_match_components.append(return_node_name)
+            type_matches.append('%s-[:`%s`]->%s' %
+                    (return_node_type_name, INSTANCE_REL, return_node_name))
+
+        model_match = ''.join(
+            itertools.ifilter(None, itertools.chain.from_iterable(
+                itertools.izip_longest(rel_match_components, node_match_components))))
+
+        matches.append('%s=(s%s)'  % (path_name, model_match))
+        matches.extend(type_matches)
+
+    return 'MATCH %s RETURN %s' % (','.join(matches), ','.join(returns))
 
 #XXX this will have to change significantly when issue #1 is worked on
 #TODO this can be broken into retrieval and rebuilding functions
@@ -251,7 +291,9 @@ def execute_select_related(models=None, query=None, index_name=None,
     """
     Retrieves select_related models and and adds them to model caches.
     """
-    if models:
+    if models is not None:
+        if len(models) == 0:
+            return
         #infer the database we're using
         model_dbs = [m.using for m in models if m.node]
         if len(set(model_dbs)) > 1:
@@ -262,7 +304,6 @@ def execute_select_related(models=None, query=None, index_name=None,
         #infer the model type
         if model_type is None:
             model_type = type(models[0])
-
         start_expr = 'node(%s)' % ','.join(str(m.id) for m in models)
         start_depth = 1
     elif index_name and query:
@@ -283,57 +324,91 @@ def execute_select_related(models=None, query=None, index_name=None,
             raise ValueError("If no fields are provided for select_related, "
                                 "max_depth must be > 0.")
         #the simple depth-only case
-        cypher_query = 'START s1 = %s '\
-                       'MATCH p1=(s1-[g*%d..%d]-r1), pt1=(t1-[:`%s`]->r1) '\
-                       'RETURN p1, r1, t1.name'
+        #TODO it looks like this only works for depth=1...
+        cypher_query = 'START s = %s '\
+                       'MATCH p0=(s-[g*%d..%d]-p0_r0), p0_r0_t-[:`%s`]->p0_r0 '\
+                       'RETURN p0, p0_r0, p0_r0_t.name'
         cypher_query %= (start_expr, start_depth, max_depth, INSTANCE_REL)
-        results = conn.cypher(cypher_query)
     elif fields:
-        #get a new start point for each field
-        starts = ['s%d=%s' % (i, start_expr) for i in xrange(len(fields))]
         #build a match pattern + type check for each field
-        match_expr = ', '.join(cypher_match_from_fields(model_type, fields))
-        return_expr = ', '.join('p%d, r%d, t%d.name' % (i,i,i)
-                                for i in xrange(len(fields)))
-        cypher_query = 'START %s '\
-                       'MATCH %s '\
-                       'RETURN %s'
-        cypher_query %= (', '.join(starts), match_expr, return_expr)
-        results = conn.cypher(cypher_query)
-
+        match_and_return_expr = cypher_from_fields(model_type, fields)
+        cypher_query = 'START s=%s %s'
+        cypher_query %= (start_expr, match_and_return_expr)
     else:
-        raise ValueError("Either a fielf list of max_depth must be provided "
-                         "for select_related.") #TODO
+        raise ValueError("Either a field list or max_depth must be provided "
+                         "for select_related.")
 
-    def get_columns(column_name_pred, table):
-        columns = [i for i,c in enumerate(table['columns']) if column_name_pred(c)]
-        col_getter = itemgetter(*columns)
-        return itertools.chain(*(rc if isinstance(rc, tuple) else (rc,)
-                                 for rc in (col_getter(r)
-                                            for r in table['data'])))
-    paths = sorted(get_columns(lambda c:c.startswith('p'), results), key=lambda p:p['length'])
-    nodes = get_columns(lambda c:c.startswith('r'), results)
-    types = get_columns(lambda c:c.startswith('t'), results)
+    results = conn.cypher(cypher_query)
 
-    #batch all relationships from paths
+    #TODO this is another example of needing a cypher generation abstraction.
+    paths = sorted(not_none(
+                     results.get_rows(lambda c:re.match('p\d+$', c) is not None)),
+                   key=lambda p:p['length'])
+    nodes, types = [], []
+    for path_i in itertools.count():
+        path_name = 'p%d' % path_i
+        if path_name not in results.column_names:
+            break
+        for node_i in itertools.count():
+            return_node_name = '%s_r%s' % (path_name, node_i)
+            return_node_type = '%s_t.name' % return_node_name
+            if not (return_node_name in results.column_names or
+                    return_node_type in results.column_names):
+                break
+            nodes = itertools.chain(nodes, results.get_rows(return_node_name))
+            types = itertools.chain(types, results.get_rows(return_node_type))
+
+    nodes = not_none(nodes)
+    types = not_none(types)
+
+    #put nodes in an id-lookup dict
+    nodes = [script_utils.LazyNode.from_dict(d) for d in nodes]
+    nodes_by_id = dict((n.id, n) for n in nodes)
+    #add any nodes we've got from the models list
+    if models is not None:
+        nodes_by_id.update(dict((m.id, script_utils.LazyNode.from_dict(m.node._dic)) for m in models))
+
+    #batch all relationships from paths and put em in a dict
     rels_by_id = {}
-    with conn.transaction():
-        for p in paths:
-            for rel_url in p['relationships']:
-                rel_id = id_from_url(rel_url)
-                if rel_id not in rels_by_id:
-                    rels_by_id[rel_id] = conn.relationships[rel_id]
+    rel_ids = []
+    for p in paths:
+        for rel_url in p['relationships']:
+            rel_ids.append(id_from_url(rel_url))
+    rels = script_utils.batch_rels(rel_ids, using)
+    for r in rels:
+        r.set_custom_node_lookup(nodes_by_id)
+        rels_by_id[r.id] = r
 
-    #build all the models
-    rel_models = (get_model(*t.split(':'))._neo4j_instance(n) for n, t in
-                    ((LazyNode(r[0]['self'], r[0]), r[1])
-                    for r in itertools.izip(nodes, types)))
+    #build all the models, ignoring types that django hasn't loaded
+    rel_nodes_types= ((n, get_model(*t.split(':')))
+                      for n, t in itertools.izip(nodes, types))
 
+    rel_models = (t._neo4j_instance(n) for n, t in rel_nodes_types if
+        (t is not None) and (t._neo4j_instance(n) not in models))
     models_so_far = dict((m.id, m) for m in itertools.chain(models, rel_models))
 
+    # TODO HACK set model rel caches to empty 
+    # in the future, we'd like to properly mark a cache as 'filled', 'empty',
+    # or 'unknown', to deal with deferred relationships versus those that have
+    # been serviced by select_related. That will require doing more bookkeeping-
+    # eg, knowing which models are at what depth in the max_depth case, and
+    # which correspond to which field in the field case.
+    # This covers the easy case, max_depth=1, and ignores the hard case of
+    # dealing with a fields list or a greater depth.
+    if fields is None and max_depth == 1:
+        for m in models:
+            for field_name, field in m._meta._relationships.items():
+                #if rel is many side
+                rel_on_model = getattr(m, field_name, None)
+                if rel_on_model is not None and hasattr(rel_on_model, '_cache'):
+                    rel_on_model._get_or_create_cache() #set the cache to loaded and empty
+                else:
+                    #otherwise single side
+                    field._set_cached_relationship(m, None)
+
     #paths ordered by shortest to longest
     paths = sorted(paths, key=lambda v:v['length'])
-    
+
     for path in paths:
         m = models_so_far[id_from_url(path['start'])]
 
@@ -350,19 +425,20 @@ def execute_select_related(models=None, query=None, index_name=None,
             field_candidates = [(k,v) for k,v in cur_m._meta._relationships.items()
                                 if str(v.rel_type)==str(rel.type) and v.direction == rel.direction]
             if len(field_candidates) < 1:
-                raise ValueError("No model field candidate for returned "
-                                    "path - there's an error in the Cypher "
-                                    "query or your model definition.")
+                continue
             elif len(field_candidates) > 1:
                 raise ValueError("Too many model field candidates for "
-                                    "returned path - there's an error in the "
-                                    "Cypher query or your model definition.")
+                                 "returned path - there's an error in the "
+                                 "Cypher query or your model definition.")
             field_name, field = field_candidates[0]
 
+            #grab the model that should fit in this part of the path
+            new_model = models_so_far[node_id]
+
             #if rel is many side
             rel_on_model = getattr(cur_m, field_name, None)
             if rel_on_model and hasattr(rel_on_model, '_cache'):
-                rel_on_model._cache.append((rel, new_model))
+                rel_on_model._add_to_cache((rel, new_model))
                 if field.ordered:
                     rel_on_model._cache.sort(
                         key=lambda r:r[0].properties.get(ORDER_ATTR, None))
@@ -371,6 +447,8 @@ def execute_select_related(models=None, query=None, index_name=None,
                 field._set_cached_relationship(cur_m, new_model)
             cur_m = new_model
 
+
+
 class Query(object):
     def __init__(self, nodetype, conditions=tuple(), max_depth=None, 
                  select_related_fields=[]):
@@ -379,9 +457,9 @@ class Query(object):
         self.max_depth = max_depth
         self.select_related_fields = []
         self.select_related = bool(select_related_fields) or max_depth
-    
+
     def add(self, prop, value, operator=OPERATORS.EXACT, negate=False):
-        #TODO validate, based on prop type, etc
+    #TODO validate, based on prop type, etc
         return type(self)(self.nodetype, conditions = self.conditions +\
                           (Condition(prop, value, operator, negate),))
 
@@ -415,6 +493,8 @@ class Query(object):
         unindexed = []
 
         for c in conditions:
+            # if c.negate:
+            #     raise NotImplementedError('Negative conditions (eg .exclude() are not supported')
             if getattr(c.field, 'id', False):
                 id_conditions.append(c)
             elif c.field.indexed:
@@ -422,7 +502,8 @@ class Query(object):
             else:
                 unindexed.append(c)
 
-        id_lookups = dict(itertools.groupby(id_conditions, lambda c: c.operator))
+        grouped_id_conds = itertools.groupby(id_conditions, lambda c: c.operator)
+        id_lookups = dict(((k, list(l)) for k, l in grouped_id_conds))
         exact_id_lookups = list(id_lookups.get(OPERATORS.EXACT, []))
         #if we have an exact lookup, do it and return
         if len(exact_id_lookups) == 1:
@@ -431,7 +512,15 @@ class Query(object):
                 node = connections[using].nodes[int(id_val)]
                 #TODO also check type!!
                 if all(matches_condition(node, c) for c in itertools.chain(indexed, unindexed)):
-                    yield self.model_from_node(node)
+                    #TODO DRY violation!
+                    model_results = [self.model_from_node(node)]
+                    if self.select_related:
+                        sel_fields = self.select_related_fields
+                        if not sel_fields: sel_fields = None
+                        execute_select_related(models=model_results,
+                                                fields=sel_fields,
+                                                max_depth=self.max_depth)
+                    yield model_results[0]
             except:
                 pass
             return
@@ -443,34 +532,45 @@ class Query(object):
         if in_id_lookups:
             id_set = reduce(and_, (set(c.value) for c in in_id_lookups))
             if id_set:
-                ext = connections[using].extensions['GremlinPlugin']
-                gremlin_script = 'g.v(%s)'
-                gremlin_script %= ','.join(str(i) for i in id_set)
-                nodes = ext.execute_script(gremlin_script)
-                #TODO also check type!!
-                for node in nodes:
-                    if all(matches_condition(node, c) for c in itertools.chain(indexed, unindexed)):
-                        yield self.model_from_node(node)
+                script = "results = Neo4Django.getVerticesByIds(ids)._();"
+                nodes = connections[using].gremlin_tx(script, ids=list(id_set))
+                ## TODO: HACKS: We don't know type coming out of neo4j-rest-client
+                #               so we check it hackily here.
+                if nodes == u'null':
+                    return
+                if hasattr(nodes, 'url'):
+                    nodes = [nodes]
+                model_results = [self.model_from_node(node) for node in nodes
+                                 if all(matches_condition(node, c) for c in itertools.chain(indexed, unindexed))]
+                #TODO DRY violation
+                if self.select_related:
+                    sel_fields = self.select_related_fields
+                    if not sel_fields: sel_fields = None
+                    execute_select_related(models=model_results,
+                                            fields=sel_fields,
+                                            max_depth=self.max_depth)
+
+                for r in model_results:
+                    yield r
                 return
             else:
-                raise ValueError('Conflicting id__in lookups - the intersection'
-                                 ' of the queried id lists is empty.')
-                                                      
+                return ## Emulates django's behavior
                                                       
         #TODO order by type - check against the global type first, so that if
-        #we get an empty result set, we can return none
+        #we get an empty result set, we can return none? this kind of impedes Q
+        #objects, though- revisit
 
-        results = {} #TODO this could perhaps be cached - think2 about it
+        results = {} #TODO this could perhaps be cached - think about it
         filtered_results = set()
 
-        #XXX: annoying workaround bc neo4jrestclient.client.Index doesn't tell equality
-        #well/properly
-        index_by_url = dict((i.url,i) for i in (c.field.index(using) for c in indexed))
+        index_by_name = dict((i.name, i) for i in (c.field.index(using) for c in indexed))
 
+        #TODO order by type
         built_q = False
-        cond_by_ind = itertools.groupby(indexed, lambda c:c.field.index(using).url)
-        for index_url, group in cond_by_ind:
-            index = index_by_url[index_url]
+        cond_by_ind = itertools.groupby(indexed, lambda c:c.field.index(using).name)
+        index_qs = []
+        for index_name, group in cond_by_ind:
+            index = index_by_name[index_name]
             q = None
             for condition in group:
                 new_q = q_from_condition(condition)
@@ -483,30 +583,27 @@ class Query(object):
                 else:
                     q = new_q
             if q is not None:
-                result_set = set(index.query(q))
-                #TODO results is currently worthless
-                results[q] = result_set
-                #TODO also needs to match at least one type, if any have been provided
-                #filter for unindexed conditions, as well
-                filtered_result = set(n for n in result_set \
-                                if all(matches_condition(n, c) for c in conditions)\
-                                    and n not in filtered_results)
-                filtered_results |= filtered_result
-                model_results = [self.model_from_node(n)
-                                 for n in filtered_result]
-                #TODO DRY violation
-                if self.select_related:
-                    sel_fields = self.select_related_fields
-                    if not sel_fields: sel_fields = None
-                    execute_select_related(index_name=index.name, query=q,
-                                           fields=sel_fields,
-                                           max_depth=self.max_depth,
-                                           model_type=self.nodetype
-                                          )
-                for r in model_results:
-                    yield r
+                index_qs.append((index_name, str(q)))
+        
+        if built_q:
+            result_set = script_utils.query_indices(index_qs, using)
 
-        if not built_q:
+            #filter for unindexed conditions, as well
+            filtered_result = set(n for n in result_set \
+                            if all(matches_condition(n, c) for c in conditions))
+            model_results = [self.model_from_node(n)
+                                for n in filtered_result]
+            #TODO DRY violation
+            if self.select_related:
+                sel_fields = self.select_related_fields
+                if not sel_fields: sel_fields = None
+                execute_select_related(models=model_results,
+                                        fields=sel_fields,
+                                        max_depth=self.max_depth)
+
+            for r in model_results:
+                yield r
+        else:
             return_filter = return_filter_from_conditions(unindexed + indexed)
             rel_types = [neo4j.Outgoing.get('<<TYPE>>'),
                          neo4j.Outgoing.get('<<INSTANCE>>')]
@@ -578,7 +675,7 @@ def condition_from_kw(nodetype, keyval):
     attname = terms[0]
     field = getattr(nodetype, attname)
     
-    if op in (OPERATORS.RANGE, OPERATORS.IN):
+    if op in (OPERATORS.RANGE, OPERATORS.IN, OPERATORS.MEMBER_IN):
         return Condition(field, tuple([field.to_neo(v) for v in keyval[1]]), op, False)
     else:
         return Condition(field, field.to_neo(keyval[1]), op, False)
@@ -611,6 +708,26 @@ class NodeQuerySet(QuerySet):
     # METHODS THAT DO DATABASE QUERIES #
     ####################################
 
+    def __getitem__(self, k):
+        """"
+        If k is a slice or there's a ._result_cache, use super __getitem__.
+        Otherwise, iterate over the queryset, loading items into the cache
+        one by one, and return last element of the cache.
+        """
+        if not isinstance(k, (int, long)) or (k < 0) or self._result_cache is not None:
+            return super(NodeQuerySet, self).__getitem__(k)
+
+        try:
+            # ._fill_cache would be handy, but doesn't work when ._iter is None
+            self._result_cache = []
+            self._iter = self._iter or self.iterator()
+            for _ in range(k + 1):
+                self._result_cache.append(next(self._iter))
+            return self._result_cache[-1]
+
+        except self.model.DoesNotExist, e:
+            raise IndexError(e.args)
+
     def iterator(self):
         using = self.db
         for model in self.query.execute(using):
@@ -632,9 +749,12 @@ class NodeQuerySet(QuerySet):
     #TODO would be awesome if this were transactional
     def get_or_create(self, **kwargs):
         try:
-            return self.get(**kwargs)
+            obj = self.get(**kwargs)
+            created = False
         except:
-            return self.create(**kwargs)
+            obj = self.create(**kwargs)
+            created = True
+        return (obj, created)
 
     @not_implemented
     def latest(self, field_name=None):
@@ -648,6 +768,8 @@ class NodeQuerySet(QuerySet):
     def delete(self):
         #TODO naive delete, should be seriously optimized- consider using
         # the batch api or some clever traversal deletion type stuff
+        #TODO When new batch delete is put in place, will need to call pre_
+        # and post_delete signals here; now they are covered in model delete()
         clone = self._clone()
         for obj in clone:
             obj.delete()
@@ -695,7 +817,7 @@ class NodeQuerySet(QuerySet):
         for c in conditions_from_kws(self.model, kwargs):
             neg_c = Condition(*c[:-1], negate=negate)
             new_query = new_query.add_cond(neg_c)
-        return self._clone(klass=NodeQuerySet, query=new_query)
+        return self._clone(query=new_query)
 
     @not_implemented
     def complex_filter(self, filter_obj):
