@@ -1,7 +1,10 @@
-
 from django.db import models
 from django.db.models.fields.related import add_lazy_relation
 from django.db.models.query_utils import DeferredAttribute
+from django.db.models.query import EmptyQuerySet
+from django.db.models.signals import post_delete
+from django.dispatch import receiver
+from django.core.exceptions import FieldError
 
 from neo4django import Incoming, Outgoing
 from neo4django.db import DEFAULT_DB_ALIAS
@@ -9,6 +12,8 @@ from neo4django.decorators import not_implemented, transactional
 from neo4django.utils import buffer_iterator, AssignableList, AttrRouter
 from neo4django.constants import INTERNAL_ATTR, ORDER_ATTR
 from base import NodeModel
+from neo4django.db.models.query import conditions_from_kws, matches_condition
+
 
 from neo4jrestclient.constants import RELATIONSHIPS_ALL, RELATIONSHIPS_IN, RELATIONSHIPS_OUT
 
@@ -35,7 +40,7 @@ class LazyModel(object):
     __target = None
 
     @property
-    def __model(self):
+    def _model(self):
         model = self.__target
         if model is None:
             raise ValueError("Lazy model not initialized!")
@@ -48,10 +53,10 @@ class LazyModel(object):
     def __getattr__(self, attr):
         if attr in ('__deepcopy__',) and self.__target is None:
             raise AttributeError
-        return getattr(self.__model, attr)
+        return getattr(self._model, attr)
 
     def __call__(self, *args, **kwargs):
-        return self.__model(*args, **kwargs)
+        return self._model(*args, **kwargs)
 
 class RelationshipBase(type):
     """
@@ -141,8 +146,9 @@ class Relationship(object):
             if rel_type.direction != direction:
                 raise ValueError("Incompatible direction!")
             rel_type = rel_type.type
+        
         self.__target = target
-        self.__name = rel_type
+        self.name = rel_type
         self.__single = single
         self.direction = direction
         self.__related_single = related_single
@@ -158,6 +164,7 @@ class Relationship(object):
     __is_reversed = False
 
     def reverse(self, target, name):
+        
         if self.direction is RELATIONSHIPS_IN:
             direction = RELATIONSHIPS_OUT
         elif self.direction is RELATIONSHIPS_OUT:
@@ -165,7 +172,7 @@ class Relationship(object):
         else:
             direction = RELATIONSHIPS_OUT
         relationship = Relationship(
-            target, rel_type=self.__name, direction=direction,
+            target, rel_type=self.name, direction=direction,
             single=self.__related_single, related_name=name,
             metadata=self.__related_meta, preserve_ordering=self.__ordered)
         relationship.__is_reversed = True
@@ -179,10 +186,7 @@ class Relationship(object):
 
     @staticmethod
     def get_name(target, single=False):
-        if single:
-            suffix = '%ss'
-        else:
-            suffix = '%s_set'
+        suffix = '%s' if single else '%s_set'
         if isinstance(target, basestring):
             name = target.rsplit('.',1)[-1]
         else:
@@ -193,20 +197,30 @@ class Relationship(object):
         if not issubclass(source, NodeModel):
             raise TypeError("Relationships may only extend from Nodes.")
         self.creation_counter = source.creation_counter
-        
+
+        # make sure this relationship doesn't overlap with another of the same
+        # type and direction
+        if hasattr(source._meta, '_relationships'):
+            for r in source._meta._relationships.values():
+                if r.rel_type == self.name and r.direction == self.direction:
+                    import warnings
+                    warnings.warn('`%s` and `%s` share a relationship type and '
+                                  'direction. Is this what you meant to do?' 
+                                  % (r.name, name))
+
         if hasattr(self, 'Model'):
             if self.__single:
                 Bound = SingleRelationship
             else:
                 Bound = MultipleRelationships
-            bound = Bound(self, source, self.__name or name, name,
+            bound = Bound(self, source, self.name or name, name,
                             self.Model)
         else:
             if self.__single:
                 Bound = SingleNode
             else:
                 Bound = MultipleNodes
-            bound = Bound(self, source, self.__name or name, name)
+            bound = Bound(self, source, self.name or name, name)
         source._meta.add_field(bound)
         if not hasattr(source._meta, '_relationships'):
             source._meta._relationships = {}
@@ -224,7 +238,7 @@ class Relationship(object):
     ###################
 
     def __getinitargs__(self):
-        return (self.__target, self.__name, self.direction, True, self.__single,
+        return (self.__target, self.name, self.direction, True, self.__single,
                 self.__related_single, self.reversed_name, self.__ordered,
                 self.__meta, self.__related_meta)
 
@@ -265,6 +279,11 @@ class BoundRelationship(AttrRouter, DeferredAttribute):
     def rel_type(self):
         return self._type
 
+    @property
+    def relationship(self):
+        return self.__rel
+
+
     def get_default(self):
         return None
 
@@ -435,7 +454,10 @@ class SingleNode(BoundRelationship):
     @transactional
     def _load_related(self, node):
         relationships = self._load_relationships(node)
-        django_relationships = filter(lambda rel: rel['_neo4django'], relationships)
+        #TODO seriously consider removing this restriction- I'm not sure I see
+        # any benefit, and it makes creating neo4django-compliant graphs that
+        # much more difficult.
+        django_relationships = filter(lambda rel: rel[INTERNAL_ATTR], relationships)
         if len(django_relationships) < 1:
             return None
         elif len(django_relationships) > 1:
@@ -542,15 +564,15 @@ class MultipleNodes(BoundRelationship):
     def _save_relationship(self, instance, node, state):
         state.__save__(node)
 
-#    @classmethod
-#    def to_neo(cls, value):
-#         pass
+   # @classmethod
+   # def to_neo(cls, value):
+   #      pass
 
-#    @classmethod
-#    def from_neo(cls, value):
-#        pass
+   # @classmethod
+   # def from_neo(cls, value):
+   #     pass
 
-    #def _set_relationship():
+    # def _set_relationship():
     #    #TODO diff rels and this list
     #    #add news rels, delete old
     #    #preserve order if we're supposed to
@@ -586,14 +608,51 @@ class RelationshipInstance(models.Manager):
     A manager that keeps state for the many side (`MultipleNodes`) of relationships.
     """
     def __init__(self, rel, obj):
-        self.__rel = rel
-        self.__obj = obj
+        self._rel = rel
+        self._obj = obj
         self._added = [] # contains domain objects
         self._removed = [] # contains relationships
         #holds cached domain objects (that have been added or loaded by query)
-        self._cache = []
+        self._cache = None
+        self._cache_unique = set([])
 
-    ordered = property(lambda self: self.__rel.ordered)
+        # sender should be the associated model (not any associated LazyModel)
+        sender = self._rel.target_model._model \
+                if hasattr(self._rel.target_model, '_model') \
+                else self._rel.target_model
+        @receiver(post_delete, sender=sender, weak=False)
+        def delete_handler(sender,**kwargs):
+            deleted_obj = kwargs.pop('instance', None)
+            if deleted_obj:
+                self._remove_from_cache(deleted_obj)
+                if deleted_obj in self._added:
+                    self._added.remove(deleted_obj)
+
+
+    ordered = property(lambda self: self._rel.ordered)
+
+    def _add_to_cache(self, *relationship_neo4j_pairs):
+        for pair in relationship_neo4j_pairs:
+            if pair not in self._cache_unique:
+                self._get_or_create_cache().append(pair)
+                self._cache_unique.add(pair)
+
+    def _remove_from_cache(self, obj):
+        if self._cache is not None:
+            for r, cached_obj in self._cache[:]:
+                if cached_obj == obj:
+                    pair = (r, cached_obj)
+                    self._cache.remove(pair)
+                    self._cache_unique.remove(pair)
+                    break
+
+    def _has_cache(self):
+        return self._cache is not None
+
+    def _get_or_create_cache(self):
+        if self._cache is None:
+            self._cache = []
+        return self._cache
 
     def __save__(self, node):
         #Deletes all relationships removed since last save and adds any new
@@ -603,19 +662,19 @@ class RelationshipInstance(models.Manager):
         for relationship in self._removed:
             relationship.delete()
         for obj in self._added:
-            new_rel = self.__rel._create_neo_relationship(node, obj)
-            self._cache.append((new_rel, obj))
+            new_rel = self._rel._create_neo_relationship(node, obj)
+            self._add_to_cache((new_rel, obj))
         self._removed[:] = []
         self._added[:] = []
 
     def _neo4j_relationships_and_models(self, node):
-        if not self._cache:
-            self._cache = [(r, self.__rel._neo4j_instance(node, r)) for r in 
-                           self.__rel._load_relationships(node, ordered=self.ordered)]
-        for tup in self._cache:
+        "Returns generator of relationship, neo4j instance tuples associated with node."
+        if self._cache is None:
+            self._add_to_cache(*[(r, self._rel._neo4j_instance(node, r)) for r in 
+                           self._rel._load_relationships(node, ordered=self.ordered)])
+        for tup in self._get_or_create_cache():
             if tup[0] not in self._removed:
                 yield tup
-
     @property
     def _new(self):
         for item in self._added:
@@ -632,7 +691,7 @@ class RelationshipInstance(models.Manager):
         the relationship, these objects will all be put at the end of the line.
         """
         for obj in objs:
-            self.__rel.accept(obj)
+            self._rel.accept(obj)
         self._added.extend(objs)
 
     def remove(self, *objs):
@@ -641,13 +700,13 @@ class RelationshipInstance(models.Manager):
         remove the first relationship to this object- otherwise, remove one
         of the relationships indiscriminately.
         """
-        rel = self.__rel
-        if hasattr(self.__obj, 'node'):
-            neo_rels = list(rel._load_relationships(self.__obj.node,
+        rel = self._rel
+        if hasattr(self._obj, 'node'):
+            neo_rels = list(rel._load_relationships(self._obj.node,
                                                     ordered=self.ordered))
             rels_by_node = defaultdict(list)
             for neo_rel in neo_rels:
-                other_end = neo_rel.start if neo_rel.end == self.__obj.node\
+                other_end = neo_rel.start if neo_rel.end == self._obj.node\
                             else neo_rel.end
                 rels_by_node[other_end].append(neo_rel)
             nodes_to_remove = [o.node for o in objs if hasattr(o, 'node')]
@@ -656,36 +715,51 @@ class RelationshipInstance(models.Manager):
                 candidate_rels = rels_by_node[obj.node]\
                         if hasattr(o, 'node') else []
                 if candidate_rels:
-                    self._removed.append(candidate_rels.pop(0))
+                    if candidate_rels[0] not in self._removed:
+                        self._removed.append(candidate_rels.pop(0))
                 else:
                     try:
                         self._added.remove(obj)
                     except ValueError:
-                        raise rel.target_model.DoesNotExist("%r is not related to %r." % (obj, self.__obj))
+                        raise rel.target_model.DoesNotExist("%r is not related to %r." % (obj, self._obj))
+                self._remove_from_cache(obj)
         else:
             for obj in objs:
                 try:
                     if obj in self._added:
                         self._added.remove(obj)
-                    else:
-                        self._cache.remove(obj)
+                    elif obj in self._cache:
+                        self._remove_from_cache(obj)
                 except ValueError:
-                    raise rel.target_model.DoesNotExist("%r is not related to %r." % (obj, self.__obj))
+                    raise rel.target_model.DoesNotExist("%r is not related to %r." % (obj, self._obj))
 
     def clear(self):
         all_objs = list(self.all())
         self.remove(*all_objs)
 
-    @not_implemented
-    def create(self, *args, **kwargs):
-        pass
+    def clone(self):
+        # Should cache be updated as well?
+        cloned = RelationshipInstance(self._rel, self._obj)
+        cloned.add(*self._new)
+        return cloned
+
+    def create(self, **kwargs):
+        kwargs[self._rel.relationship._related_name] = self._obj
+        new_model = self._rel.relationship.target_model(**kwargs)
+        # TODO: saving twice, should only need
+        # to save self._obj after #89 fix
+        new_model.save()
+        self._obj.save()
 
     @not_implemented
     def get_or_create(self, *args, **kwargs):
         pass
 
     def get_query_set(self):
-        return RelationshipQuerySet(self, self.__rel, self.__obj)
+        return RelationshipQuerySet(self, self._rel, self._obj)
+
+    def get_empty_query_set(self):
+        return EmptyQuerySet()
 
 class RelationshipQuerySet(object):
     def __init__(self, inst, rel, obj):
@@ -693,6 +767,25 @@ class RelationshipQuerySet(object):
         self.__rel = rel
         self.__obj = obj
 
+    def filter(self, **kwargs):
+        "Returns RelationshipQuerySet with filtered items"
+        inst = self.__inst
+        added = list(inst._new)
+
+        new_inst = self.__inst.clone()
+        new_inst.clear()
+        if added:
+            iterable = added
+        else:
+            # TODO: is wrapping in getattr necessary?
+            node = getattr(self.__obj, 'node', None)
+            iterable = [i for r, i in inst._neo4j_relationships_and_models(node)] if node else []
+
+        for item in iterable:
+            if any(matches_condition(item.node, c) for c in conditions_from_kws(self.__rel.relationship.target_model, kwargs)):
+                new_inst.add(item)
+        return RelationshipQuerySet(new_inst, self.__rel, self.__obj)
+
     def __saved_instances(self, node):
         for rel, item in self.__inst._neo4j_relationships_and_models(node):
             if self.__keep_relationship(rel) and self.__keep_instance(item):
@@ -712,6 +805,20 @@ class RelationshipQuerySet(object):
             if self.__keep_instance(item):
                 yield item
 
+    def __len__(self):
+        return sum(1 for _ in self)
+
+    # Taken from Django's QuerySet repr
+    def __repr__(self):
+        REPR_OUTPUT_SIZE = 4
+        data = list(self[:REPR_OUTPUT_SIZE + 1])
+        if len(data) > REPR_OUTPUT_SIZE:
+            data[-1] = "...(remaining elements truncated)..."
+        return repr(data)
+
+    def __getitem__(self, key):
+        return list(self)[key]
+
     def __keep_instance(self, obj):
         return True # TODO: filtering
 
@@ -721,3 +828,10 @@ class RelationshipQuerySet(object):
     @not_implemented
     def get(self, **lookup):
         pass
+
+    def delete(self):
+        for obj in self:
+            obj.delete()
+
+    def count(self):
+        return len(self)
