@@ -3,6 +3,7 @@ from nose.tools import with_setup, eq_
 from django.core import exceptions
 
 from time import time
+import itertools
 import sys, datetime
 stdout = sys.stdout
 
@@ -12,22 +13,16 @@ def setup():
            DEFAULT_DB_ALIAS, return_filter_from_conditions, Condition, models,\
            RelatedCat, RelatedDog 
 
-    from neo4django.tests import Person, neo4django, gdb, models
-    from neo4django.db import DEFAULT_DB_ALIAS
+    from neo4django.tests import Person, neo4django, gdb
+    from neo4django.db import DEFAULT_DB_ALIAS, models
     from neo4django.db.models.query import Query, OPERATORS, \
             return_filter_from_conditions, Condition
 
-    class IndexedMouse(models.NodeModel):
-        name = models.StringProperty(indexed=True)
-        age = models.IntegerProperty(indexed=True)
+    from django.db.models import get_model
 
-    class RelatedCat(models.NodeModel):
-        name = models.StringProperty()
-        chases = models.Relationship(IndexedMouse, rel_type='chases')
-
-    class RelatedDog(models.NodeModel):
-        name = models.StringProperty()
-        chases = models.Relationship(RelatedCat, rel_type='chases')
+    RelatedCat = get_model('tests','RelatedCat')
+    RelatedDog = get_model('tests','RelatedDog')
+    IndexedMouse = get_model('tests','IndexedMouse')
 
 def teardown():
     gdb.cleandb()
@@ -55,7 +50,7 @@ def test_delete():
         pass
     else:
         raise AssertionError("Jack's pk is still in the graph- he wasn't "
-                             "created.")
+                             "deleted.")
 
 @with_setup(None, teardown)
 def test_iter():
@@ -86,12 +81,18 @@ def test_dates():
     assert paper.name == 'Papes'
     assert other.name == 'other'
     assert paper.datetime < other.datetime
-    
+
+def test_none():
+    eq_(len(Person.objects.none()), 0)
+
+def make_mice(names, ages):
+    for name, age in zip(names, ages):
+        IndexedMouse.objects.create(name=name,age=age)
+
 mouse_names = ['jerry','Brain', 'Pinky']
 mouse_ages = [2,3,2]
 def setup_mice():
-    for name, age in zip(mouse_names, mouse_ages):
-        IndexedMouse.objects.create(name=name,age=age)
+    make_mice(mouse_names, mouse_ages)
 
 def make_people(names, ages):
     pairs = zip(names, ages)
@@ -268,7 +269,7 @@ def test_in_id():
     uninteresting_man = Person.objects.create(name=boring_name, age=boring_age)
 
     Person.objects.create(age=boring_age)
-
+    
     people = list(Person.objects.filter(id__in=(interesting_man.id, uninteresting_man.id)))
     eq_(len(people), 2)
     eq_([boring_age, age], sorted(p.age for p in people))
@@ -278,6 +279,29 @@ def test_in_id():
     eq_(len(people), 1)
     eq_(people[0].id, uninteresting_man.id)
 
+    single_person = list(Person.objects.filter(id__in=(interesting_man.id,)))
+    eq_(len(single_person), 1)
+
+    no_people = list(Person.objects.filter(id__in=(1000,)))
+    eq_(len(no_people), 0)
+
+    # Test chaining
+    only_interesting = Person.objects.filter(
+        id__in=(interesting_man.id,)
+        ).filter(id__in=(interesting_man.id, uninteresting_man.id))
+    eq_(len(only_interesting), 1)
+
+    only_interesting = Person.objects.filter(
+        id__in=(interesting_man.id,)
+        ).filter(id__in=(uninteresting_man.id,))
+    eq_(len(only_interesting), 0)
+    
+    # Passing in an empty qs -- replicate django
+    eq_(len(Person.objects.filter(id__in=[])), 0)
+
+    # Passing in qs with None -- replicate django
+    eq_(len(Person.objects.filter(id__in=[uninteresting_man.id, None])), 1)
+
 def setup_teens():
     setup_people()
     make_people(['Tina', 'Rob', 'Tiny Tim'], [13, 15, 12])
@@ -319,7 +343,7 @@ def test_filter_lte():
 alphabet = [chr(i + 97) for i in range(26)]
 def test_filter_range():
     import random
-    ages_and_names = zip(*[(random.sample(alphabet, 6), i + 70) for i in xrange(20)])
+    ages_and_names = zip(*[(''.join(random.sample(alphabet, 6)), i + 70) for i in xrange(20)])
     make_people(*ages_and_names)
     octogenarians = Person.objects.filter(age__range=(80, 89))
     assert all(80 <= p.age <= 89 for p in octogenarians), "These guys aren't all in their 80's!"
@@ -379,6 +403,26 @@ def test_filter_in():
     eq_(len(q), 4)
     assert all(p.age in [15,12] for p in q)
 
+@with_setup(None, teardown)
+def test_filter_array_member_in():
+    """
+    Tests the `field__member_in` array membership field lookup.
+    """
+    class TooManyAccounts(Person):
+        emails = models.StringArrayProperty(indexed=True)
+
+    emails = ['test1@example.com','test2@example.com','test3@example.com']
+    p1 = TooManyAccounts.objects.create(emails=emails[:2])
+    p2 = TooManyAccounts.objects.create(emails=emails[1:])
+
+    q_1only = TooManyAccounts.objects.filter(emails__member_in=[emails[0]])
+    q_both = TooManyAccounts.objects.filter(emails__member_in=emails)
+
+    eq_(len(q_1only), 1)
+    eq_(list(q_1only)[0].id, p1.id)
+
+    eq_(set(p.id for p in q_both), set((p1.id, p2.id)))
+
 #test isnull
 
 @with_setup(None, teardown)
@@ -407,6 +451,14 @@ def test_in_bulk():
     eq_(people[interesting_man.id].name, name)
     eq_([boring_age, age], sorted(p.age for p in people.values()))
 
+@with_setup(setup_people, teardown)
+def test_in_bulk_not_found():
+    """
+    Tests QuerySet.in_builk() with items not found.
+    """
+    people = Person.objects.in_bulk([999999])
+    eq_(people, {})
+
 @with_setup(setup_mice_and_people, teardown)
 def test_contains():
     q1 = Person.objects.filter(name__contains='a')
@@ -471,3 +523,67 @@ def test_select_related():
     
     #try the hierarchy with a field-based select_related
     check_dog_hier_from_q(RelatedDog.objects.all().select_related('chases','chases__chases'))
+
+@with_setup(None, teardown)
+def test_large_query():
+    ages = range(1, 151)
+    names = ['a mouse'] * len(ages)
+    make_mice(names, ages)
+
+    mice =  list(IndexedMouse.objects.filter(age__in=ages))
+    eq_(len(mice), len(ages))
+
+@with_setup(None, teardown)
+def test_zerovalued_lookup():
+    ages = range(2)
+    make_mice(['a','a'], ages)
+
+    mice =  list(IndexedMouse.objects.filter(age__in=ages))
+    eq_(len(mice), len(ages))
+
+@with_setup(None, teardown)
+def test_get_or_create():
+    name = "Kristian"
+    (obj1, created) = Person.objects.get_or_create(name=name)
+    assert created == True, "Person should have been created but wasn't"
+    (obj2, created) = Person.objects.get_or_create(name=name)
+    assert created == False, "Person should not have been created, one should already exist"
+    assert obj1 == obj2, "Second get_or_create() should have returned Person created by first get_or_create()"
+
+def test_object_index():
+    class PollIdx(models.NodeModel):
+            question = models.StringProperty()
+            def __unicode__(self):
+                return self.question
+    p1 = PollIdx(question="Who's the best")
+    p2 = PollIdx(question="How's the weather?")
+    p3 = PollIdx(question="What's up?")
+    p1.save()
+    p2.save()
+    p3.save()
+
+    eq_(len(PollIdx.objects.all()), 3)
+    p0 = PollIdx.objects.all()[0]
+    p1 = PollIdx.objects.all()[1]
+    p2 = PollIdx.objects.all()[2]
+    assert len(set([p0, p1, p2, p0])) == 3, "There should be 3 different polls"
+
+    qsall = PollIdx.objects.all()
+    # Should fill up cache one by one
+    eq_(qsall._result_cache, None)
+    p0 = qsall[0]
+    eq_(len(qsall._result_cache), 1)
+    p1 = qsall[1]
+    eq_(len(qsall._result_cache), 2)
+    p2 = qsall[2]
+    eq_(len(qsall._result_cache), 3)
+    assert len(set([p0, p1, p2, p0, p1])) == 3, "There should be 3 different polls"
+
+    qsall = PollIdx.objects.all()
+    # Filling up the cache first
+    len(qsall)
+    eq_(len(qsall._result_cache), 3)
+    p0 = qsall[0]
+    p1 = qsall[1]
+    p2 = qsall[2]
+    assert len(set([p0, p1, p2, p0, p1, p2])) == 3, "There should still be 3 different polls"
